// Utils.C: utility functions

// system includes
#include <math.h>
#include <stdlib.h>
#include <assert.h>
#include <iostream.h>
#include <strstream.h>

// GL includes
#include <GL/gl.h>

// local includes
#include "SCtypes.h"
#include "ivwraplib/SaveWindow.h"

// min two floats
float fmin(float a, float b)
{
  return (a < b) ? a : b;
}

// max two floats
float fmax(float a, float b)
{
  return (a > b) ? a : b;
}

// return random float in [low..high]
float randr(float low, float high)
{
  return(low + (rand() / ((1<<15)-1.0)) * (high-low));
}

// max component of reflectance
float vmax(Color color)
{
  return fmax (color[0], fmax (color[1], color[2]));
}

// multiply two vectors, componentwise, to produce vector
Vector vmult(Vector &a, Vector &b)
{
  return Vector (a[0] * b[0], a[1] * b[1], a[2] * b[2]);
}

// perpendicular point/plane distance
float distance(Plane &H, Point &p)
{
  return H[0] * p[0] + H[1] * p[1] + H[2] * p[2] + H[3];
}

// adjust H.d to contain p
int plane_from_three_points(Point &a, Point &b, Point &c, Plane &H)
{
  Vector n = (b - a).cross(c - a);
  if (n.length() < 1.0E-6F)
    return 0;
  n.normalize();
  H = Plane(n[0], n[1], n[2], -n.dot(a));
  return 1;
}

// adjust H.d to contain p
void plane_from_point_normal(Point &p, Vector &n, Plane &H)
{
  H = Plane(n[0], n[1], n[2], -n.dot(p));
}

// construct plane containing p, parallel to par, perpendicular to perp
int plane_from_point_perp_par(Point &p, Vector &par,
			      Vector &perp, OUTPUT Plane &H)
{
  // H's normal is perp
  Vector normal = perp.cross(par);
  if (normal.length() < 1.0E-6F)
    return 0;

  // adjust H.d to contain p
  normal.normalize();
  plane_from_point_normal(p, normal, H);
  return 1;
}

void saveWindow (unsigned long *pixels, int width, int height, int inspect)
{
  static char fname[128];
  static int fnum = 0;
  ostrstream filename(fname, sizeof fname);

  filename << "ivscan" << (int)getpid() << ".";
  filename.fill('0');
  filename.width(4);
  filename << fnum++ << ".rgb" << ends;
  savebuffer(fname, pixels, width, height);
  if (inspect) {
	char cbuffer[128], *pwd;
	ostrstream command(cbuffer, sizeof cbuffer);
	if (pwd = getenv("PWD"))
	  command << "cd " << pwd << "; \n\t\t";
	command << "/usr/sbin/ipaste " << fname << ends;
	cout << "saveWindow(): executing '" << cbuffer << "'" << endl;
	system(cbuffer);
  }
}

GLuint ABGR_to_RGBA(GLuint abgr)
{
  return (((abgr >> 0) & 0xff) << 24) |    // R
         (((abgr >> 8) & 0xff) << 16) |    // G
	 (((abgr >> 16) & 0xff) << 8) |    // B
	 (((abgr >> 24) & 0xff) << 0);     // A
}

void setPixelDelayed (int x, int y, SbColor color, Environment *env)
{
  // convert float vector to packed rgba format
  GLuint rgba = ABGR_to_RGBA (color.getPackedValue());
  //GLuint rgba = color.getPackedValue();
 
  // store in local framebuffer
  env->pixelData[x + y * env->avpSize[X]] = rgba;

  // if reached end of raster, write raster to window
  if (x == env->avpSize[X] - 1) {
    glRasterPos2i(0,  y);
    glDrawPixels(env->avpSize[X], 1, GL_RGBA, GL_UNSIGNED_BYTE,
		  &env->pixelData[x + y * env->avpSize[X]]);
  }
}

// Draw pixel to front buffer in immediate mode.
void setPixelImmed(int x, int y, SbColor color, Environment *env)
{
  // convert float vector to packed rgba format
  // GLuint rgba = ABGR_to_RGBA(color.getPackedValue());
  GLuint rgba = color.getPackedValue();

  // store in local framebuffer
  env->pixelData[x + y * env->avpSize[X]] = rgba;

  // write pixel to window
  glRasterPos2i(x, y);
  glDrawPixels(1, 1, GL_RGBA, GL_UNSIGNED_BYTE,  &rgba);

  // wait for pipeline to finish
  glFlush();
}

// Draw pixel, blown up to NxN!
static void setBlownUpPixel(int x, int y, Color color, Environment *env, 
			    int blowup)
{
  GLuint rgba = color.getPackedValue();
  int bhalf = blowup/2;

  for ( int i = -(bhalf-1); i <= bhalf-1; i++ ) 
    for ( int j = -(bhalf-1); j <= bhalf-1; j++ ) {  
      // store in local framebuffer
      env->pixelData[(bhalf + blowup * x + i) +
                     (bhalf + blowup * y + j) * env->avpSize[X]] = rgba;

      // write pixel to window
      glRasterPos2i( bhalf + blowup * x + i, bhalf + blowup * y + j);
      glDrawPixels( 1, 1, GL_RGBA, GL_UNSIGNED_BYTE, &rgba );
      }

  // wait for pipeline to finish
  glFlush();
}

// Draw pixel, blown up to 4x4!
void setPixel4(int x, int y, Color color, Environment *env)
{
  setBlownUpPixel (x,y,color,env, 4);
}

// Draw pixel, blown up to 8x8!
void setPixel8(int x, int y, SbColor color, Environment *env)
{
  setBlownUpPixel (x,y,color,env, 8);
}

// Draw pixel, blown up to 16x16!
void setPixel16(int x, int y, SbColor color, Environment *env)
{
  setBlownUpPixel (x,y,color,env,16);
}

// Draw pixel, blown up to 32x32!
void setPixel32(int x, int y, SbColor color, Environment *env)
{
  setBlownUpPixel (x,y,color,env,32);
}

// Draw pixel to local framebuffer, but not to screen
void setPixelBuffer(int x, int y, SbColor color, Environment *env)
{
  // convert float vector to packed rgba format
  // GLuint rgba = ABGR_to_RGBA(color.getPackedValue());
	GLuint rgba = color.getPackedValue();

  // store in local framebuffer
  env->pixelData[x + y * env->avpSize[X]] = rgba;
}


Color CompMult (Color c1, Color c2)
{
  return Color (c1[0] * c2[0], c1[1] * c2[1], c1[2] * c2[2]);
}


// Draw line from x1,y1 to x2,y2 onto the pixel buffer of env.
void drawLine (int x1, int y1, int x2, int y2, 
	       SbColor color, Environment *env)
{

// drawLine (int x1, int y1, int x2, int y2);
// handles all 8 octants, all special cases
int x, y, dx, dy, npix;
int xinc, yinc, xup, yup; // advance step; error step
int e, einc, emax;        // error; err inc; err limit

dx = x2-x1;  dy = y2-y1;  // x,y deltas and sign bits
xinc = (dx >= 0) ? 1 : -1;
yinc = (dy >= 0) ? 1 : -1;

if ( (xinc * dx) > (yinc * dy) ) { // loop across x
  npix = (dx * xinc) + 1;          // note:  non-neg
  emax = dx * xinc;
  einc = (dy * yinc) << 1;
  xup = 0;
  yup = yinc;
  yinc = 0; // adv, err steps
  }
else {                             // loop across y
  npix = (dy * yinc) + 1;          // note:  non-neg
  emax = dy * yinc;
  einc = (dx * xinc) << 1;
  yup = 0;
  xup = xinc;
  xinc = 0; // adv, err steps
  }

x = x1;
y = y1;
e = 0;    // error initially zero
while ( npix-- > 0 ) {
  setPixelImmed(x, y, color, env);
  x += xinc;
  y += yinc; 
  e += einc; // advance step
  if ( e > emax ) {
    x += xup;
    y += yup;
    e -= emax << 1; // error step
    }
  }
} //end drawLine
